/*
** tifftopnm.c - converts a Tagged Image File to a portable anymap
**
** Derived by Jef Poskanzer from tif2ras.c, which is:
**
** Copyright (c) 1990 by Sun Microsystems, Inc.
**
** Author: Patrick J. Naughton
** naughton@wind.sun.com
**
** Permission to use, copy, modify, and distribute this software and its
** documentation for any purpose and without fee is hereby granted,
** provided that the above copyright notice appear in all copies and that
** both that copyright notice and this permission notice appear in
** supporting documentation.
**
** This file is provided AS IS with no warranties of any kind.  The author
** shall have no liability with respect to the infringement of copyrights,
** trade secrets or any patents by this file or any part thereof.  In no
** event will the author be liable for any lost revenue or profits or
** other special, indirect and consequential damages.
*/

/* Implementation note:

  This probably ought to be reimplemented using TIFFRGBAImageGet() 
  from the Tiff library, because that routine automatically decodes
  anything and everything that might be in a Tiff image file now or
  in the future.

  Much of the code below is basically copied from the Tiff library
  because we use the more primitive TIFFReadScanLine() access method.

  -Bryan 00.03.05
*/

#define _BSD_SOURCE 1
    /* Make sure strdup() is in string.h */

#include <string.h>
#include "pnm.h"
#ifdef VMS
#ifdef SYSV
#undef SYSV
#endif
#include <tiffioP.h>
#endif
#ifdef DJGPP
#include <tiffioP.h>
#endif
#include <tiffio.h>

/* The following are in current tiff.h, but so that we can compile against
   older tiff libraries, we define them here.
*/

#ifndef PHOTOMETRIC_LOGL
#define PHOTOMETRIC_LOGL 32844
#endif
#ifndef PHOTOMETRIC_LOGLUV
#define PHOTOMETRIC_LOGLUV 32845
#endif

#define MAXCOLORS 1024
#ifndef PHOTOMETRIC_DEPTH
#define PHOTOMETRIC_DEPTH 32768
#endif

/* The following describe the structure of the raster buffer */
static int maxval;
static unsigned short fillorder;
static unsigned short bps, spp;

/* The NEXTSAMPLE macro effects the reading of a sample from the raster
   buffer via the above cursor into the variable 'sample'.
*/
struct cmdline_info {
    /* All the information the user supplied in the command line,
       in a form easy for the program to use.
    */
    char *input_filename;
    int headerdump;
    char *alpha_filename;
    int alpha_stdout;
} cmdline;



static void
parse_command_line(int argc, char ** argv,
                   struct cmdline_info *cmdline_p);

static void read_directory(TIFF * const tif,
                           unsigned short * const fillorder_p,
                           unsigned short * const bps_p,
                           unsigned short * const spp_p,
                           unsigned short * const photomet_p,
                           unsigned short * const planarconfig_p,
                           unsigned int * const cols_p,
                           unsigned int * const rows_p,
                           const int headerdump) {
/*----------------------------------------------------------------------------
   Read various values of TIFF tags from the TIFF directory, and default 
   them if not in there.  Exit program with error message if required tags
   aren't there or values are inconsistent or beyond our capabilities.
   if 'headerdump' is true, issue informational messages about what we find.
-----------------------------------------------------------------------------*/

    int rc;
    unsigned short tiff_bps;
    unsigned short tiff_spp;
    unsigned short tiff_fillorder;

    if (headerdump)
        TIFFPrintDirectory(tif, stderr, TIFFPRINT_NONE);

    rc = TIFFGetField(tif, TIFFTAG_BITSPERSAMPLE, &tiff_bps);
    *bps_p = (rc == 0) ? 1 : tiff_bps;

    if (*bps_p < 1 || (*bps_p > 8 && *bps_p != 16))
        pm_error("This program can process Tiff images with only "
                 "1-8 or 16 bits per sample.  The input Tiff image "
                 "has %d bits per sample.", bps);

    rc = TIFFGetField(tif, TIFFTAG_FILLORDER, &tiff_fillorder);
    /* I don't know if it is legal for a Tiff file not to have a 
       FILLORDER tag, but until I added 16 bit Tiff capability in November
       2000, we never tried to read the tag.  I don't want to break this 
       program for older image inputs by erroring out on a missing FILLORDER
       tag here.  -Bryan
       */
    *fillorder_p = (rc == 0) ? FILLORDER_MSB2LSB : tiff_fillorder;

    rc = TIFFGetField(tif, TIFFTAG_SAMPLESPERPIXEL, &tiff_spp);
    *spp_p = (rc == 0) ? 1: tiff_spp;

    rc = TIFFGetField(tif, TIFFTAG_PHOTOMETRIC, photomet_p);
    if (rc == 0)
        pm_error("PHOTOMETRIC tag is not in Tiff file.  "
                 "TIFFGetField() of it failed.\n"
                 "This means the input is not valid Tiff.");

    if (*spp_p > 1) {
        rc = TIFFGetField(tif, TIFFTAG_PLANARCONFIG, planarconfig_p);
        if (rc == 0)
            pm_error("PLANARCONFIG tag is not in Tiff file, though it "
                     "has more than one sample per pixel.  "
                     "TIFFGetField() of it failed.  This means the input "
                     "is not valid Tiff.");
    } else {
        *planarconfig_p = PLANARCONFIG_CONTIG;
    }

    switch (*spp_p) {
	case 1:
	case 3:
	case 4:
        break;
	default:
        pm_error(
            "This program can only handle 1-channel gray scale "
            "or 1- or 3-channel color.  The input Tiff file has %d channels "
            "(samples per pixel)",
            *spp_p);
	}

    switch(*planarconfig_p) {
	case PLANARCONFIG_CONTIG:
	    break;
	case PLANARCONFIG_SEPARATE:
	    if (*photomet_p != PHOTOMETRIC_RGB && 
            *photomet_p != PHOTOMETRIC_SEPARATED)
            pm_error("This program can handle separate planes only "
                     "with RGB (PHOTOMETRIC tag = %d) or SEPARATED "
                     "(PHOTOMETRIC tag = %d) data.  The input Tiff file " 
                     "has PHOTOMETRIC tag = %d.",
                     PHOTOMETRIC_RGB, PHOTOMETRIC_SEPARATED, &photomet_p);
	    break;
	default:
	    pm_error("Unrecognized PLANARCONFIG tag value in Tiff input: %d.\n",
                 *planarconfig_p);
	}

    rc = TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, cols_p);
    if (rc == 0)
        pm_error("Input Tiff file is invalid.  It has no IMAGEWIDTH tag.");
    rc = TIFFGetField( tif, TIFFTAG_IMAGELENGTH, rows_p );
    if (rc == 0)
        pm_error("Input Tiff file is invalid.  It has no IMAGELENGTH tag.");

    if (headerdump) {
        pm_message( "%dx%dx%d image", *cols_p, *rows_p, *bps_p * *spp_p );
        pm_message( "%d bits/sample, %d samples/pixel", *bps_p, *spp_p );
	}
}



static void
readscanline(TIFF * const tif, unsigned short * const samplebuf,
             unsigned char scanbuf[], const int row, const int plane,
             const unsigned int cols, const unsigned short bps, 
             const unsigned short fillorder) {
/*----------------------------------------------------------------------------
   Read one scanline out of the Tiff input and store it into samplebuf[].
   Unlike the scanline returned by the Tiff library function, samplebuf[]
   is composed of one sample per array element, which makes it easier for
   our caller to process.

   scanbuf[] is a scratch array for our use, which is big enough to hold
   a Tiff scanline.
-----------------------------------------------------------------------------*/
    int rc;
    const unsigned int bpsmask = (1 << bps) - 1;
      /* A mask for taking the lowest 'bps' bits of a number */

    rc = TIFFReadScanline(tif, scanbuf, row, plane);
    if (rc < 0)
        pm_error( "Unable to read row %d, plane %d of input Tiff image.  "
                  "TIFFReadScanline() failed.",
                  row, plane);
    else if (bps == 8) {
        int sample;
        for (sample = 0; sample < cols*spp; sample++) 
            samplebuf[sample] = scanbuf[sample];
    } else if (bps < 8) {
        /* Note that in this format, samples do not span bytes.  Rather,
           each byte may have don't-care bits in the low order positions.
           At least that's how I infer the format from reading pnmtotiff.c
           -Bryan 00.11.18
           */
        int sample;
        int bitsleft;
        unsigned char * inP;

        bitsleft = 0;
        inP = scanbuf;
        for (sample = 0; sample < cols*spp; sample++) {
            if (bitsleft == 0) {
                ++inP; 
                bitsleft = 8;
            } 
            bitsleft -= bps; 
            samplebuf[sample] = (*inP >> bitsleft) & bpsmask; 
        }
    } else if (bps == 16) {
        /* Here's an idea to make this faster:  determine the endianness of
           the machine we're running on, and if it's the same as that of the
           input ('fillorder'), just do a memcpy().
           */
        switch (fillorder) {
        case FILLORDER_MSB2LSB: {
            int sample;
            unsigned char * inP;
            inP = scanbuf;  /* initial value */
            for (sample = 0; sample < cols*spp; sample++) {
                samplebuf[sample] = (*inP << 8) | *(inP+1);
                inP += 2;
            }
        }
        break;
        case FILLORDER_LSB2MSB: {
            int sample;
            unsigned char * inP;
            inP = scanbuf;  /* initial value */
            for (sample = 0; sample < cols*spp; sample++) {
                samplebuf[sample] = (*(inP+1) << 8) | *inP;
                inP += 2;
            }
        }
        break;
        default:
            pm_error("Internal error: invalid fillorder passed to "
                     "readscanline()");
        }
    } else
        pm_error("Internal error: invalid bits per sample passed to "
                 "readscanline()");
}



static void
pick_cmyk_pixel(const unsigned short samplebuf[], const int sample_cursor,
                xelval * const r_p, xelval * const b_p, xelval * const g_p) {

    unsigned char c, y, m, k;

    c = samplebuf[sample_cursor+0];
    y = samplebuf[sample_cursor+1];
    m = samplebuf[sample_cursor+2];
    k = samplebuf[sample_cursor+3];

    /* I don't know why this works.  It seems to me that yellow pigment should
       translate into both red and green light, not just green as you see here.
       But this is the formula used by TIFFRGBAImageGet() in the Tiff library.
       */
    
    *r_p = ((255-k)*(255-c))/255;
    *g_p = ((255-k)*(255-y))/255;
    *b_p = ((255-k)*(255-m))/255;
}


int
main(int argc, char * argv[]) {

    unsigned int cols, rows;
    int grayscale, format;
    int numcolors;
    register TIFF* tif;
    FILE *alpha_file, *imageout_file;
    int row, i;
    register int col;
    unsigned char * scanbuf;
    unsigned short * samplebuf;
    xel* xelrow;
        /* The ppm-format row of the image row we are presently converting */
    gray* alpharow;
        /* The pgm-format row representing the alpha values for the image 
           row we are presently converting.
           */
    xel colormap[MAXCOLORS];
    unsigned char sample;
    unsigned short photomet, planarconfig, inkset;
    unsigned short* redcolormap;
    unsigned short* greencolormap;
    unsigned short* bluecolormap;

    pnm_init( &argc, argv );

    parse_command_line(argc, argv, &cmdline);

    if ( cmdline.input_filename != NULL ) {
        tif = TIFFOpen( cmdline.input_filename, "r" );
        if ( tif == NULL )
            pm_error( "error opening TIFF file %s", cmdline.input_filename );
	} else {
        tif = TIFFFdOpen( 0, "Standard Input", "r" );
        if ( tif == NULL )
            pm_error( "error opening standard input as TIFF file" );
	}

    if (cmdline.alpha_stdout)
        alpha_file = stdout;
    else if (cmdline.alpha_filename == NULL) 
        alpha_file = NULL;
    else {
        alpha_file = pm_openw(cmdline.alpha_filename);
    }

    if (cmdline.alpha_stdout) 
        imageout_file = NULL;
    else
        imageout_file = stdout;

    read_directory(tif, &fillorder, &bps, &spp, &photomet, &planarconfig, 
                   &cols, &rows, 
                   cmdline.headerdump);

    if ( bps == 1 && spp == 1 ) {
        if ( cmdline.headerdump )
            pm_message("monochrome" );
        grayscale = 1;
        maxval = 1;
	} else {
        /* How come we don't deal with the photometric for the monochrome 
           case (make sure it's one we know)?  -Bryan 00.03.04
           */
        switch ( photomet ) {
          case PHOTOMETRIC_MINISBLACK:
            if ( cmdline.headerdump )
              pm_message( "%d graylevels (min=black)", maxval + 1 );
            grayscale = 1;
            maxval = (1 << bps) - 1;
            break;

          case PHOTOMETRIC_MINISWHITE:
            if ( cmdline.headerdump )
              pm_message( "%d graylevels (min=white)", maxval + 1 );
            grayscale = 1;
            maxval = (1 << bps) - 1;
            break;

          case PHOTOMETRIC_PALETTE:
            if ( cmdline.headerdump )
              pm_message( "colormapped" );
            if ( ! TIFFGetField( tif, TIFFTAG_COLORMAP, 
                                &redcolormap, &greencolormap, &bluecolormap ) )
              pm_error( "error getting colormaps" );
            numcolors = 1 << bps;
            if ( numcolors > MAXCOLORS )
                pm_error( "too many colors" );
            maxval = PNM_MAXMAXVAL;
            grayscale = 0;
            for ( i = 0; i < numcolors; ++i ) {
                register xelval r, g, b;
                r = (long) redcolormap[i] * PNM_MAXMAXVAL / 65535L;
                g = (long) greencolormap[i] * PNM_MAXMAXVAL / 65535L;
                b = (long) bluecolormap[i] * PNM_MAXMAXVAL / 65535L;
                PPM_ASSIGN( colormap[i], r, g, b );
            }
            break;

          case PHOTOMETRIC_SEPARATED:
            if ( cmdline.headerdump )
              pm_message( "color separation" );
            if ( TIFFGetField( tif, TIFFTAG_INKNAMES, &inkset ) == 1
                && inkset != INKSET_CMYK)
            if (inkset != INKSET_CMYK) 
               pm_error("This color separation file uses an inkset (%d) ",
                        "we can't handle.  We handle only CMYK.", inkset);
            if (spp != 4) 
              pm_error("This CMYK color separation file is %d samples per "
                       "pixel.  We need 4 samples, though: C, M, Y, and K.  ",
                       spp);
            grayscale = 0;
            maxval = (1 << bps) - 1;
            break;
            
          case PHOTOMETRIC_RGB:
            if ( cmdline.headerdump )
              pm_message( "RGB truecolor" );
            grayscale = 0;
            maxval = (1 << bps) - 1;
            break;

          case PHOTOMETRIC_MASK:
            pm_error( "don't know how to handle PHOTOMETRIC_MASK" );

          case PHOTOMETRIC_DEPTH:
            pm_error( "don't know how to handle PHOTOMETRIC_DEPTH" );

          case PHOTOMETRIC_YCBCR:
            pm_error( "don't know how to handle PHOTOMETRIC_YCBCR" );

          case PHOTOMETRIC_CIELAB:
            pm_error( "don't know how to handle PHOTOMETRIC_CIELAB" );

          case PHOTOMETRIC_LOGL:
            pm_error( "don't know how to handle PHOTOMETRIC_LOGL" );

          case PHOTOMETRIC_LOGLUV:
            pm_error( "don't know how to handle PHOTOMETRIC_LOGLUV" );

          default:
            pm_error( "unknown photometric: %d", photomet );
	    }
	}
    if ( maxval > PNM_OVERALLMAXVAL )
        pm_error("bits/sample (%d) in the input image is too large.\n",
                 bps);
    if ( grayscale )
	{
	if ( maxval == 1 )
	    {
	    format = PBM_TYPE;
	    pm_message( "writing PBM file" );
	    }
	else
	    {
	    format = PGM_TYPE;
	    pm_message( "writing PGM file" );
	    }
	}
    else
	{
	format = PPM_TYPE;
	pm_message( "writing PPM file" );
	}

    scanbuf = (unsigned char *) malloc(TIFFScanlineSize(tif));
    if (scanbuf == NULL)
        pm_error("can't allocate memory for scanline buffer");

    samplebuf = (unsigned short *) malloc(cols * sizeof(unsigned short) * spp);
    if (samplebuf == NULL)
        pm_error ("can't allocate memory for row buffer");

    if (imageout_file != NULL) 
        pnm_writepnminit( imageout_file, 
                          cols, rows, (xelval) maxval, format, 0 );
    if (alpha_file != NULL) 
        pgm_writepgminit( alpha_file, cols, rows, (gray) maxval, 0 );

    xelrow = pnm_allocrow( cols );
    alpharow = pgm_allocrow( cols );

    for ( row = 0; row < rows; ++row ) {
        /* Read one row of samples into samplebuf[] */
        readscanline(tif, samplebuf, scanbuf, row, 0, 
                     cols, bps, fillorder);

        switch ( photomet ) {
          case PHOTOMETRIC_MINISBLACK:
            for ( col = 0; col < cols; ++col ) {
                PNM_ASSIGN1( xelrow[col], samplebuf[col] );
                alpharow[col] = 0;
            }
            break;

          case PHOTOMETRIC_MINISWHITE:
            for ( col = 0; col < cols; ++col ) {
                sample = maxval - sample;
                PNM_ASSIGN1( xelrow[col], samplebuf[col] );
                alpharow[col] = 0;
            }
            break;

          case PHOTOMETRIC_PALETTE:
            for ( col = 0; col < cols; ++col ) {
                /* We know the following array index is in bounds because
                   we filled samplebuf with samples of 'bps' bits each and
                   we verified that the largest number that fits in 'bps'
                   bits is less than MAXCOLORS, the dimension of the array.
                   */
                xelrow[col] = colormap[samplebuf[col]];
                alpharow[col] = 0;
            }
            break;

          case PHOTOMETRIC_SEPARATED:
            if (planarconfig == PLANARCONFIG_CONTIG) {
                int col, sample;
                for ( col = 0, sample = 0; col < cols; ++col, sample+=spp) {
                    xelval r, g, b;
                    pick_cmyk_pixel( samplebuf, sample, &r, &b, &g );
            
                    PPM_ASSIGN( xelrow[col], r, g, b );
                    alpharow[col] = 0;
                }
            } else 
                pm_error("Can't handle multiple plane color separation file");
            break;

          case PHOTOMETRIC_RGB:
            if( planarconfig == PLANARCONFIG_CONTIG ) {
                int col, sample;
                for ( col = 0, sample = 0; col < cols; ++col, sample+=spp ) {
                    PPM_ASSIGN( xelrow[col], samplebuf[sample+0],
                                samplebuf[sample+1], samplebuf[sample+2] );
                    if (spp < 4)
                        alpharow[col] = samplebuf[sample+3];
                    else
                        alpharow[col] = 0;
                }
            } else {
                /* The input is in separate planes, so the scan line in the 
                   buffer now is just the red.  We need to read twice more
                   to get the green and blue.
                   */
                int col;
                
                /* First clear the value and assign the reds */
                for ( col = 0; col < cols; ++col ) {
                    PPM_ASSIGN( xelrow[col], 0, 0, 0 );
                    PPM_PUTR( xelrow[col], samplebuf[col] );
                }
                
                /* Next the greens */
                readscanline( tif, samplebuf, scanbuf, row, 1,
                             cols, bps, fillorder );
                for ( col = 0; col < cols; ++col ) 
                    PPM_PUTG( xelrow[col], samplebuf[col] );

                /* And finally the blues */
                readscanline( tif, samplebuf, scanbuf, row, 2,
                             cols, bps, fillorder );
                for ( col = 0; col < cols; ++col ) 
                    PPM_PUTB( xelrow[col], samplebuf[col] );
                /* Could there be an alpha plane?  (We assume no.  But if so,
                   here is where to read it) */
                for ( col = 0; col < cols; ++col) 
                    alpharow[col] = 0;
            }		
            break;
            
          default:
            pm_error( "internal error:  unknown photometric in the picking "
                     "routine: %d", photomet );
        }
        if (imageout_file != NULL) 
            pnm_writepnmrow( imageout_file, 
                             xelrow, cols, (xelval) maxval, format, 0 );
        if (alpha_file != NULL) 
            pgm_writepgmrow( alpha_file, alpharow, cols, (gray) maxval, 0);
    }

    if (imageout_file != NULL) 
        pm_close( imageout_file );
    if (alpha_file != NULL)
        pm_close( alpha_file );

    /* If the program failed, it previously aborted with nonzero completion
       code, via various function calls.
    */
    return 0;
    }



static void
parse_command_line(int argc, char ** argv,
                   struct cmdline_info *cmdline_p) {
/*----------------------------------------------------------------------------
   Note that many of the strings that this function returns in the
   *cmdline_p structure are actually in the supplied argv array.  And
   sometimes, one of these strings is actually just a suffix of an entry
   in argv!
-----------------------------------------------------------------------------*/
    optStruct *option_def = malloc(100*sizeof(optStruct));
        /* Instructions to OptParseOptions on how to parse our options.
         */
    unsigned int option_def_index;

    option_def_index = 0;   /* incremented by OPTENTRY */
    OPTENTRY('h', "headerdump", OPT_FLAG,   &cmdline_p->headerdump,     0);
    OPTENTRY(0,   "alphaout",   OPT_STRING, &cmdline_p->alpha_filename, 0);

    /* Set the defaults */
    cmdline_p->headerdump = 0;
    cmdline_p->alpha_filename = NULL;

    optParseOptions(&argc, argv, option_def, 0);
        /* Uses and sets argc, argv, and all of *cmdline_p. */

    if (argc - 1 == 0)
        cmdline_p->input_filename = NULL;  /* he wants stdin */
    else if (argc - 1 == 1)
        cmdline_p->input_filename = strdup(argv[1]);
    else 
        pm_error("Too many arguments.  The only argument accepted\n"
                 "is the input file specificaton");

    if (cmdline_p->alpha_filename && 
        strcmp(cmdline_p->alpha_filename, "-") == 0 )
        cmdline_p->alpha_stdout = 1;
    else 
        cmdline_p->alpha_stdout = 0;
}

